import 'package:flutter/material.dart';
import 'package:flutter_riverpod/flutter_riverpod.dart';

class TodoExample extends StatelessWidget {
  const TodoExample({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return const ProviderScope(
      child: TodoList(),
    );
  }
}

class TodoList extends ConsumerStatefulWidget {
  const TodoList({Key? key}) : super(key: key);

  @override
  ConsumerState<TodoList> createState() => _TodoListState();
}

class _TodoListState extends ConsumerState<TodoList> {
  bool _isUncheckedFilter = false;

  @override
  Widget build(BuildContext context) {
    final todos = ref.watch(todosProvider.select((tds) => tds.whereTodos(_isUncheckedFilter)));

    return Column(
      children: [
        TodoListFilter(
          filter: _isUncheckedFilter,
          onFilterChanged: _onFilterChanged,
        ),
        const Divider(height: 1),
        Expanded(
          child: ListView.builder(
            itemBuilder: (_, index) => TodoListItem(
              todos[index],
              onCheckedChanged: (value) => _onCheckedChanged(todos[index], value),
            ),
            itemCount: todos.length,
          ),
        ),
      ],
    );
  }

  void _onFilterChanged(bool value) {
    setState(() => _isUncheckedFilter = value);
  }

  void _onCheckedChanged(Todo todo, bool? value) {
    final updatedTodo = todo.copyWith(checked: value != null && value);
    ref.read(todosProvider.notifier).updateTodo(updatedTodo);
  }
}

class TodoListFilter extends StatelessWidget {
  const TodoListFilter({
    Key? key,
    required this.filter,
    required this.onFilterChanged,
  }) : super(key: key);

  final bool filter;
  final ValueChanged<bool> onFilterChanged;

  @override
  Widget build(BuildContext context) {
    return Padding(
      padding: const EdgeInsets.symmetric(horizontal: 16),
      child: Row(
        mainAxisAlignment: MainAxisAlignment.spaceBetween,
        children: [
          const Text("Unchecked"),
          Switch(
            value: filter,
            onChanged: onFilterChanged,
          ),
        ],
      ),
    );
  }
}

class TodoListItem extends StatelessWidget {
  const TodoListItem(
    this.todo, {
    Key? key,
    this.onCheckedChanged,
  }) : super(key: key);

  final Todo todo;
  final ValueChanged<bool?>? onCheckedChanged;

  @override
  Widget build(BuildContext context) {
    return CheckboxListTile(
      value: todo.checked,
      onChanged: onCheckedChanged,
      title: Text(
        todo.title,
        overflow: TextOverflow.ellipsis,
      ),
      controlAffinity: ListTileControlAffinity.leading,
    );
  }
}

final todosProvider = StateNotifierProvider<TodosStateNotifier, List<Todo>>(//
    (ref) => TodosStateNotifier(ref.watch(todosRepositoryProvider)));

class TodosStateNotifier extends StateNotifier<List<Todo>> {
  TodosStateNotifier(this.todosRepository) : super([]) {
    load();
  }

  final TodosRepository todosRepository;

  void load() {
    state = todosRepository.getTodos();
  }

  void updateTodo(Todo updatedTodo) {
    state = todosRepository.updateTodos(updatedTodo);
  }
}

final todosRepositoryProvider = Provider<TodosRepository>((ref) => FakeTodosRepository(todos));

abstract class TodosRepository {
  List<Todo> getTodos();

  List<Todo> updateTodos(Todo updatedTodo);
}

class FakeTodosRepository implements TodosRepository {
  FakeTodosRepository(this.todos);

  List<Todo> todos;

  @override
  List<Todo> getTodos() {
    return todos;
  }

  @override
  List<Todo> updateTodos(Todo updatedTodo) {
    return todos = todos.copyWithTodo(updatedTodo);
  }
}

const todos = [
  Todo(
    id: 0,
    title: 'Unit test passed',
    checked: true,
  ),
  Todo(
    id: 1,
    title: 'Code reviewed',
    checked: true,
  ),
  Todo(
    id: 2,
    title: 'Acceptance criteria for each issue met',
  ),
  Todo(
    id: 3,
    title: 'Functional tests passed',
  ),
  Todo(
    id: 4,
    title: 'Non-functional requirements met',
  ),
  Todo(
    id: 5,
    title: 'Product owner accepts the User Story',
  ),
];

class Todo {
  const Todo({
    required this.id,
    required this.title,
    this.checked = false,
  });

  final int id;
  final String title;
  final bool checked;

  Todo copyWith({
    int? id,
    String? title,
    bool? checked,
  }) {
    return Todo(
      id: id ?? this.id,
      title: title ?? this.title,
      checked: checked ?? this.checked,
    );
  }
}

extension TodoListExtension on List<Todo> {
  List<Todo> whereTodos(bool isUnchecked) => //
      where((todo) => !isUnchecked || !todo.checked).toList();

  List<Todo> copyWithTodo(Todo updatedTodo) => //
      [for (Todo todo in this) todo.id == updatedTodo.id ? updatedTodo : todo];
}
